/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.AWTEvent;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.event.AWTEventListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferStrategy;
import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.swing.GroupLayout;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.LayoutStyle;

import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.ChunkSelectionTracker;
import se.llbit.chunky.world.ChunkUpdateTracker;
import se.llbit.chunky.world.ChunkView;
import se.llbit.chunky.world.DeleteChunksJob;
import se.llbit.chunky.world.NullWorld;
import se.llbit.chunky.world.World;
import se.llbit.chunky.world.WorldRenderer;

/**
 * This class is a bit messy and should be split into a monitor/viewer/controller.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
@SuppressWarnings("serial")
public class Chunky extends Canvas {
	
	public static final int DEFAULT_SCALE = 8;
	
	private BufferStrategy strategy;
	private World world = new NullWorld();
	public JFrame frame;
	
	private static final double ORIGIN_X = 0.5;
	private static final double ORIGIN_Z = 0.5;
	private double x0 = ORIGIN_X;
	private double z0 = ORIGIN_Z;

	private Minimap minimap;

	private int blockScale = DEFAULT_SCALE;

	private Controls controls;
	private int currentDimension = 0;
	private int rx;
	private int ry;
	private int rw;
	private int rh;
	private boolean selectRect = false;
	private RenderBuffer renderBuffer;
	private Chunk.Renderer chunkRenderer = Chunk.surfaceRenderer;
	private ChunkUpdateTracker chunkUpdateTracker = new ChunkUpdateTracker();
	private WorldRenderer worldRenderer = new WorldRenderer(chunkUpdateTracker);
	private ChunkSelectionTracker chunkSelection = new ChunkSelectionTracker();


	public static final String getAppName() {
		return String.format("%s %s (%s)", //$NON-NLS-1$
			Messages.getString("Chunky.appname"), //$NON-NLS-1$
			Messages.getString("Chunky.version"), //$NON-NLS-1$
			Messages.getString("Chunky.revision")); //$NON-NLS-1$
	}

	/**
	 * Create a new instance of the application GUI.
	 * @param args 
	 */
	public Chunky(String[] args) {
	    chunkSelection.addChunkUpdateListener(chunkUpdateTracker);
	    
		frame = new JFrame(getAppName());
		minimap = new Minimap(this);
		controls = new Controls(this, minimap);
		setMinimumSize(new Dimension(10, 10));
		setPreferredSize(new Dimension(800, 600));
		setBackground(Color.white);
		setIgnoreRepaint(true);
		setupFrame();
		
		URL url = getClass().getResource(Messages.getString("Chunky.icon")); //$NON-NLS-1$
		if (url != null)
			frame.setIconImage(Toolkit.getDefaultToolkit().getImage(url));
		
		createBufferStrategy(2);
		strategy = getBufferStrategy();
		
		minimap.init();
		
		InputListener listener = new InputListener();
		frame.addComponentListener(listener);
		this.addMouseListener(listener);
		this.addMouseMotionListener(listener);
		this.addMouseWheelListener(listener);
		
		Toolkit.getDefaultToolkit().addAWTEventListener(listener, AWTEvent.KEY_EVENT_MASK);
		
		requestFocus();
		
		boolean selectedWorld = false;
		if (args.length > 0) {
		    File worldDir = new File(args[0]);
		    if (worldDir.exists() && worldDir.isDirectory()) {
		        loadWorld(new World(worldDir));
		        selectedWorld = true;
		    }
		}
		
		if (!selectedWorld)
		    new WorldSelector(this);
	}
	
	/**
	 * Set up the main application frame.
	 */
	private void setupFrame() {
		GroupLayout layout = new GroupLayout(frame.getContentPane());
		frame.getContentPane().setLayout(layout);
		layout.setHorizontalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
				.addComponent(this, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(controls, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
		);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addComponent(controls, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
			.addComponent(this, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
		);

		frame.pack();
		
		ChunkView view = new ChunkView();
		renderBuffer = new RenderBuffer(view);
		
		frame.setResizable(true);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setLocationByPlatform(true);
		frame.setVisible(true);
	}

	public synchronized void loadWorld(World newWorld) {
		world.dispose();
		world.removeChunkDeletionListener(chunkSelection);
		world.removeChunkUpdateListener(chunkUpdateTracker);
		
		world = newWorld;
		world.addChunkDeletionListener(chunkSelection);
		world.addChunkUpdateListener(chunkUpdateTracker);
		
		chunkSelection.clearSelection();
		
		x0 = ORIGIN_X;
		z0 = ORIGIN_Z;
		world.loadWorld(currentDimension);
		controls.setLayer(world.currentLayer());
		if (world.havePlayerPos()) {
			x0 = world.playerPosX() / 16;
			z0 = world.playerPosZ() / 16;
		}
		renderBuffer.flushCache();
		minimap.flushCache();
		updateView();
		refresh();
		controls.setPlayerY(world.playerLocY());
		controls.enableDimension(0, world.haveDimension(0));
		controls.enableDimension(-1, world.haveDimension(-1));
		controls.enableDimension(1, world.haveDimension(1));
	}
	
	public synchronized int getScale() {
		return blockScale;
	}
	
	public synchronized void updateView() {
		world.setView(getView());
	}

	public static void main(String[] args) {
		long max = Runtime.getRuntime().maxMemory();
		if (max / (1024*1024) < 900) {
			// TODO: Warn user about low heap limit!
			// pop up a dialog to ask if a .bat or .sh should be created
			// to start Chunky with more memory
		}
		Chunky thisObj = new Chunky(args);
		thisObj.refresh();
		thisObj.refreshLoop();
	}

	/**
	 * Periodically check for changes in the world state and
	 * refresh the view if necessary.
	 */
	private void refreshLoop() {
		try {
			while (true) {
				if (chunkUpdateTracker.haveUpdatedChunks())
					refresh();
				else if (world.mapUpdated())
					refreshMinimap();
				Thread.sleep(125);
			}
		} catch (InterruptedException e) {
		}
	}

	/**
	 * Refresh/redraw the view.
	 */
	public synchronized void refresh() {
		ChunkView view = getView();
		
		Graphics g = strategy.getDrawGraphics();
		renderBuffer.setView(view,chunkRenderer, world.currentLayer());
		worldRenderer.render(world, renderBuffer, chunkRenderer);
		g.clearRect(0, 0, view.width, view.height);
		renderBuffer.renderBuffered(g);
		worldRenderer.renderHUD(world, chunkSelection, g, renderBuffer, chunkRenderer, true);
		drawSelectionRect(g);
		g.dispose();
		strategy.show();
		refreshMinimap();
	}

	private ChunkView getView() {
		int width = getWidth();
		int height = getHeight();
		double mw = width / (32.0 * blockScale);
		double mh = height / (32.0 * blockScale);
		return new ChunkView(x0 - mh, x0 + mh, z0 - mw, z0 + mw, width, height, 16 * blockScale);
	}

	private void refreshMinimap() {
		minimap.refresh(worldRenderer, world, x0, z0, renderBuffer.getView());
	}

	private void drawSelectionRect(Graphics g) {
		if (selectRect) {
			g.setColor(Color.red);
			g.drawRect(rx, ry, rw, rh);
		}
	}

	public class InputListener implements MouseListener, MouseMotionListener, ComponentListener,
		MouseWheelListener, KeyListener, AWTEventListener {
		private int ox = 0;
		private int oy = 0;
		private boolean ctrlModifier = false;
		private boolean shiftModifier = false;
		private boolean dragging = false;
		public InputListener() {
		}
		private void setOrigin(MouseEvent e) {
			ox = e.getX();
			oy = e.getY();			
		}
		private void onDrag(MouseEvent e) {
			int dx = ox - e.getX();
			int dy = oy - e.getY();
			
			if (dx == 0 && dy == 0)
				return;
			
			if (shiftModifier) {
				setSelectionRect(ox, oy, -dx, -dy);
				updateView();
				refresh();
			} else {
				if (!dragging && Math.abs(dx) < 5 && Math.abs(dy) < 5)
					return;
				dragging = true;
				setOrigin(e);
				adjustWorldOrigin(dx, dy);
				updateView();
				refresh();
			}
		}
		@Override
		public void mouseClicked(MouseEvent e) {
		}
		@Override
		public void mouseEntered(MouseEvent e) {
		}
		@Override
		public void mouseExited(MouseEvent e) {
		}
		@Override
		public void mousePressed(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON3) {
			    //new Chunk3DView(Chunky.this, getChunkBlockData(e.getX(), e.getY()));
				//getChunkInfo(e.getX(), e.getY());
			} else
				setOrigin(e);
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON3)
				return;
			
			if (!shiftModifier) {
				if (!dragging)
					selectChunk(e.getX(), e.getY());
				
			} else {
				selectWithinRect();
				clearSelectionRect();
			}
			dragging = false;
		}
		@Override
		public void mouseDragged(MouseEvent e) {
			onDrag(e);
		}
		@Override
		public void mouseMoved(MouseEvent e) {
		}
		@Override
		public void componentHidden(ComponentEvent arg0) {
		}
		@Override
		public void componentMoved(ComponentEvent arg0) {
			refresh();
		}
		@Override
		public void componentResized(ComponentEvent arg0) {
			updateView();
			refresh();
		}
		@Override
		public void componentShown(ComponentEvent arg0) {
			refresh();
		}
		@Override
		public void mouseWheelMoved(MouseWheelEvent e) {
			int diff = e.getWheelRotation();
			
			if (!ctrlModifier) {
				setLayer(getLayer() + diff);
				controls.setLayer(getLayer());
			} else {
				setScale(getScale() - diff);
				controls.setScale(getScale());
			}
		}
		@Override
		public void keyPressed(KeyEvent e) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_CONTROL:
				ctrlModifier = true;
				break;
			case KeyEvent.VK_SHIFT:
				shiftModifier = true;
				break;
			case KeyEvent.VK_DELETE:
				deleteSelectedChunks();
				break;
			}
		}
		@Override
		public void keyReleased(KeyEvent e) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_CONTROL:
				ctrlModifier = false;
				break;
			case KeyEvent.VK_SHIFT:
				shiftModifier = false;
				clearSelectionRect();
				break;
			}
		}
		@Override
		public void keyTyped(KeyEvent e) {
		}
		@Override
		public void eventDispatched(AWTEvent event) {
			if (event instanceof KeyEvent) {
				KeyEvent e = (KeyEvent) event;
				if (e.getID() == KeyEvent.KEY_PRESSED) {
					keyPressed(e);
				} else if (e.getID() == KeyEvent.KEY_RELEASED) {
					keyReleased(e);
				}
			}
		}
	}

	public synchronized void setRenderer(Chunk.Renderer renderer) {
		this.chunkRenderer = renderer;
		updateView();
		refresh();
	}

    public synchronized void selectWithinRect() {
		if (selectRect) {
			int width = getWidth();
			int height = getHeight();
			int cx0 = (int) Math.floor(x0 + (ry - height/2) / (blockScale * 16.0));
			int cz0 = (int) Math.floor(z0 + (width/2 - rx - rw) / (blockScale * 16.0));
			int cx1 = (int) Math.floor(x0 + (ry - height/2 + rh) / (blockScale * 16.0));
			int cz1 = (int) Math.floor(z0 + (width/2 - rx) / (blockScale * 16.0));
			chunkSelection.selectChunks(world, cx0, cz0, cx1, cz1);
			controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		}
	}

	public void clearSelectionRect() {
		if (selectRect) {
			selectRect = false;
			refresh();
		}
	}

	public void setSelectionRect(int ox, int oy, int dx, int dy) {
		rx = Math.min(ox, ox+dx);
		ry = Math.min(oy, oy+dy);
		rw = Math.abs(dx);
		rh = Math.abs(dy);
		selectRect = true;
	}

	public synchronized void selectChunk(int x, int y) {
	    int cx = (int) Math.floor(x0 + (y - getHeight()/2) / (double)(blockScale*16));
        int cz = (int) Math.floor(z0 - (x - getWidth()/2) / (double)(blockScale*16));
		chunkSelection.selectChunk(world, cx, cz);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}
	
	public synchronized void getChunkInfo(int x, int y) {
		int cx = (int) Math.floor(x0 + (y - getHeight()/2) / (double)(blockScale*16));
		int cz = (int) Math.floor(z0 - (x - getWidth()/2) / (double)(blockScale*16));
		ChunkPosition pos = ChunkPosition.get(cx, cz);
		JOptionPane.showMessageDialog(null, "chunk clicked: "+pos);
	}
	

    public List<byte[]> getChunkBlockData(int ix, int iy) {
        int cx = (int) Math.floor(x0 + (iy - getHeight()/2) / (double)(blockScale*16));
        int cz = (int) Math.floor(z0 - (ix - getWidth()/2) / (double)(blockScale*16));
        
        List<byte[]> blockData = new ArrayList<byte[]>();
        for (int z = 0; z < RenderController.CHUNKS_X; ++z) {
            for (int x = 0; x < RenderController.CHUNKS_X; ++x) {
                ChunkPosition pos = ChunkPosition.get(cx+x, cz+z);
                blockData.add(world.getChunk(pos).getBlockData(world.getDataDirectory()));        
            }
        }
        return blockData;
    }

	/**
	 * Adjust the world origin according to a change in pixel space of dx, dy.
	 * 
	 * @param dx
	 * @param dy
	 */
	public synchronized void adjustWorldOrigin(int dx, int dy) {
		x0 += dy / (double)(blockScale*16);
		z0 -= dx / (double)(blockScale*16);
	}

	public synchronized void setScale(int value) {
		int scaleNew = Math.max(1, Math.min(16, value));
		if (scaleNew != blockScale) {
			blockScale = scaleNew;
			updateView();
			refresh();
		}
	}
	
	public synchronized void setLayer(int value) {
		int layerNew = Math.max(0, Math.min(Chunk.Y_MAX-1, value));
		if (layerNew != world.currentLayer()) {
			world.setCurrentLayer(layerNew);
			if (chunkRenderer == Chunk.layerRenderer)
				renderBuffer.flushCache();
			else
				minimap.flushCache();
			updateView();
		}
	}
	
	public synchronized int getLayer() {
		return world.currentLayer();
	}

	public void deleteSelectedChunks() {
		Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
				Messages.getString("Chunky.AcceptDelete_lbl")}; //$NON-NLS-1$
		int n = JOptionPane.showOptionDialog(null,
				Messages.getString("Chunky.DeleteDialog_msg"), //$NON-NLS-1$
				Messages.getString("Chunky.DeleteDialog_title"), //$NON-NLS-1$
				JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE,
				null,
				options,
				options[0]);
		if (n == 1) {

		    Collection<ChunkPosition> selected = chunkSelection.getSelection();
		    ProgressPanel progress = controls.getProgressPanel();
	        if (!selected.isEmpty() && !progress.isBusy()) {
	            DeleteChunksJob job = new DeleteChunksJob(world, selected, progress);
	            job.start();
	        }
		}
	}

	public void setDimension(int value) {
		if (value != currentDimension) {
			currentDimension = value;
			loadWorld(world);
		}
	}

	public synchronized void clearSelectedChunks() {
		chunkSelection.clearSelection();
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}

	public static File getSavesDirectory() {
		String home = System.getProperty("user.home", ".");   //$NON-NLS-1$ //$NON-NLS-2$
		String os = System.getProperty("os.name").toLowerCase();  //$NON-NLS-1$
		
		if (os.contains("win")) {  //$NON-NLS-1$
			String appdata = System.getenv("APPDATA");  //$NON-NLS-1$
			if (appdata != null)
				return new File(appdata, ".minecraft/saves/");  //$NON-NLS-1$
			else
				return new File(home, ".minecraft/saves/");  //$NON-NLS-1$
		} else if (os.contains("mac")) {  //$NON-NLS-1$
			return new File(home, "Library/Application Support/minecraft/saves/");  //$NON-NLS-1$
		} else {
			return new File(home, ".minecraft/saves/");  //$NON-NLS-1$
		}
	}
	
	public Color getHighlightColor() {
		return worldRenderer.getHighlightColor();
	}
	
	public Block getHighlightBlock() {
		return worldRenderer.getHighlightBlock();
	}

	public void setHighlightEnable(boolean value) {
		worldRenderer.setHighlightEnabled(value);
		renderBuffer.flushCache();
		refresh();
	}
	
	public boolean isHighlightEnabled() {
		return worldRenderer.isHighlightEnabled();
	}

	public void highlightBlock(Block hlBlock) {
		worldRenderer.highlightBlock(hlBlock);
		// TODO: make change sensitive
		renderBuffer.flushCache();
		refresh();
	}

	public void setHighlightColor(Color newColor) {
		worldRenderer.setHighlightColor(newColor);
		// TODO: make change sensitive
		renderBuffer.flushCache();
		refresh();
	}

	public synchronized void moveView(int dx, int dz) {
		x0 += dx;
		z0 += dz;
		updateView();
		refresh();
	}

	public void selectAll() {
		chunkSelection.selectAll(world);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		refresh();
	}

	public void invertSelection() {
		chunkSelection.invertSelection(world);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		refresh();
	}

	public File getWorldDirectory() {
		return world.getWorldDirectory();
	}

	public synchronized void exportZip(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
		    if (targetFile.exists()) {
		        Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
		                Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
		        int n = JOptionPane.showOptionDialog(null,
		                String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
		                        targetFile.getName()),
		                Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
		                JOptionPane.YES_NO_OPTION,
		                JOptionPane.WARNING_MESSAGE,
		                null,
		                options,
		                options[0]);
		        if (n != 1)
		            return;
		    }
			new ZipExportJob(world, chunkSelection.getSelection(), targetFile, progress).start();
		}
	}

	public void renderView(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
		    if (targetFile.exists()) {
		        Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
		                Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
		        int n = JOptionPane.showOptionDialog(null,
		                String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
		                        targetFile.getName()),
		                Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
		                JOptionPane.YES_NO_OPTION,
		                JOptionPane.WARNING_MESSAGE,
		                null,
		                options,
		                options[0]);
		        if (n != 1)
		            return;
		    }
		    
		    if (progress.tryStartJob()) {
		    	progress.setJobName("PNG export");
		    	progress.setJobSize(1);
		    	renderBuffer.renderPng(targetFile);
		    	progress.finishJob();
		    }
		}
	}
	
	public void renderDimension(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
		    if (targetFile.exists()) {
		        Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
		                Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
		        int n = JOptionPane.showOptionDialog(null,
		                String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
		                        targetFile.getName()),
		                Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
		                JOptionPane.YES_NO_OPTION,
		                JOptionPane.WARNING_MESSAGE,
		                null,
		                options,
		                options[0]);
		        if (n != 1)
		            return;
		    }
		    
		    if (world.numChunks() != 0) {
		    	// don't render PNG if world is empty
		    	new RenderDimensionJob(world, targetFile, progress).start();
		    } else {
		    	System.err.println("No chunks to render!");
		    }
		}
	}

}
