/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.util.List;

import javax.vecmath.Matrix3f;
import javax.vecmath.Vector3f;

import se.llbit.chunky.meta.Color;
import se.llbit.chunky.meta.Ray;
import se.llbit.chunky.meta.Ray.RayPool;
import se.llbit.chunky.meta.VectorPool;
import se.llbit.chunky.octree.Octree;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.BlockType;

public class RenderController extends Thread {

	private static final int NUM_WORKERS = 4;
	private Canvas3D canvas;
	private Thread[] workers = new Thread[NUM_WORKERS];
	private BufferedImage[] buffers = new BufferedImage[NUM_WORKERS];
	private int canvasWidth;
	private int canvasHeight;
	private int[] width = new int[NUM_WORKERS];
	private int[] height = new int[NUM_WORKERS];
	private int[] xoffset = new int[NUM_WORKERS];
	private int[] yoffset = new int[NUM_WORKERS];
	
	private Octree octree;
	
	/**
	 * Bitmask for worker threads.
	 */
	private int tickets;
	
	/**
	 * Bitmask for done tiles.
	 */
	private int done;
	
	public static final int CHUNKS_X = 5;
	
    private Vector3f cameraPos = new Vector3f(0, 80, 0);
    private Vector3f up = new Vector3f(0, 1, 0);
    
    private boolean renderSlow = false;
    
    private Matrix3f bufferedTransform = new Matrix3f();
    private Vector3f bufferedCameraPos = new Vector3f();
    private boolean bufferedRenderSlow;
    
    private int frameCounter = 0;
    private Matrix3f cameraTransform = new Matrix3f();
    
    private boolean refresh = true;
    private float yaw = 0;
    private float pitch = 0;
    private RayPool[] rayPool;
    private VectorPool[] vectorPool;
	
	public RenderController(Canvas3D canvas, List<byte[]> blockData) {
		super("Render Controller");
		
		updateTransforms();
		
		octree = new Octree(15);
        for (int x = 0; x < CHUNKS_X; ++x) {
            for (int z = 0; z < CHUNKS_X; ++z) {
                byte[] data = blockData.get(x + z*CHUNKS_X);
                for (int cy = 0; cy < 256; ++cy) {
                    for (int cz = 0; cz < 16; ++cz) {
                        for (int cx = 0; cx < 16; ++cx) {
                        	int type = 0xFF & data[cx + 16 * (cz + 16 * cy)];
                        	if (BlockType.water.contains(type))
                        		type = Block.WATER.id;
                            octree.set(type,
                            		cx + x*16 - CHUNKS_X * 8, cy, cz + z*16 - CHUNKS_X * 8);
                        }
                    }
                }
            }
        }
        
		this.canvas = canvas;
		
		rayPool = new RayPool[NUM_WORKERS];
		vectorPool = new VectorPool[NUM_WORKERS+1];
		
		// start worker threads
		for (int i = 0; i < NUM_WORKERS; ++i) {
		    rayPool[i] = new Ray.RayPool();
		    vectorPool[i] = new VectorPool();
		    workers[i] = new Render3DWorker(this, i);
		    workers[i].start();
		}
		vectorPool[NUM_WORKERS] = new VectorPool();
	}
	
	public synchronized void setYaw(int i) {
		if (yaw == i)
			return;
		yaw = i;
		updateTransforms();
	}
	
	public synchronized void setPitch(int i) {
        if (pitch == i)
            return;
        pitch = i;
        updateTransforms();
    }
    
	public synchronized void setZoom(int i) {
    }
    
	private synchronized void waitOnRefreshRequest() throws InterruptedException {
		while (!refresh)
			wait();
		refresh = false;
	}
	
	public synchronized void refresh() {
	    refresh = true;
	    notifyAll();
	}
	
	public void run() {
		try {
			while (!isInterrupted()) {

				waitOnRefreshRequest();
				
				long frameTime = System.currentTimeMillis();
				updateBuffers();
				giveTickets();
				waitOnWorkers();
				updateCanvas();
				System.out.println(String.format("frame %d took %.3fs",
				        frameCounter++,
				        (System.currentTimeMillis() - frameTime) / 1000.f));
			}
		} catch (InterruptedException e) {
		}
	}
	
	private void updateBuffers() {
		if (canvasWidth != canvas.getWidth() || canvasHeight != canvas.getHeight()) {
			canvasWidth = canvas.getWidth();
			canvasHeight = canvas.getHeight();
			
			width[0] = canvasWidth/2;
			height[0] = canvasHeight/2;
			buffers[0] = new BufferedImage(width[0], height[0], BufferedImage.TYPE_4BYTE_ABGR);
			xoffset[0] = yoffset[0] = 0;
			
			width[1] = canvasWidth - canvasWidth/2;
			height[1] = canvasHeight/2;
			buffers[1] = new BufferedImage(width[1], height[1], BufferedImage.TYPE_4BYTE_ABGR);
			xoffset[1] = canvasWidth/2;
			yoffset[1] = 0;
			
			width[2] = canvasWidth/2;
            height[2] = canvasHeight - canvasHeight/2;
            buffers[2] = new BufferedImage(width[2], height[2], BufferedImage.TYPE_4BYTE_ABGR);
            xoffset[2] = 0;
            yoffset[2] = canvasHeight/2;
            
            width[3] = canvasWidth - canvasWidth/2;
            height[3] = canvasHeight - canvasHeight/2;
            buffers[3] = new BufferedImage(width[3], height[3], BufferedImage.TYPE_4BYTE_ABGR);
            xoffset[3] = canvasWidth/2;
            yoffset[3] = canvasHeight/2;
		}
		
		bufferedTransform.set(cameraTransform);
		bufferedCameraPos.set(cameraPos);
		bufferedRenderSlow = renderSlow;
		renderSlow = false;
	}
	
	private synchronized void updateCanvas() {
		Graphics g = canvas.getGraphics();
		for (int i = 0; i < NUM_WORKERS; ++i) {
			g.drawImage(buffers[i], xoffset[i], yoffset[i], null);
		}
		g.dispose();
		canvas.update();
	}
	
	private synchronized void waitOnWorkers() throws InterruptedException {
		while (done != (1<<NUM_WORKERS) - 1)
			wait();
		done = 0;
	}
	
	private synchronized void giveTickets() {
		tickets = (1<<NUM_WORKERS) - 1;
		notifyAll();
	}

	public final synchronized void workerWait(int workerId) throws InterruptedException {
		while ((tickets & (1<<workerId)) == 0)
			wait();
		tickets = tickets ^ (1<<workerId);
	}
	
	public final synchronized void workerDone(int workerId) {
		done = done | (1<<workerId);
		notifyAll();
	}
	
	public final void work(int workerId, float[] buf0, float[] buf1, float[] buf2) {

		BufferedImage buffer = buffers[workerId];
    	int canvasWidth = canvas.getWidth();
    	int canvasHeight= canvas.getHeight();
    	
        Vector3f light = vectorPool[workerId].get(4000, 10000, 4000);

        // TODO FOV
        Vector3f o = vectorPool[workerId].get(this.bufferedCameraPos);
        float aspect = canvasWidth / (float) canvasHeight;
        for (int x = 0; x < width[workerId]; ++x) {
            for (int y = 0; y < height[workerId]; ++y) {
                Vector3f d = vectorPool[workerId].get(
                        -1 + (2.f * (y + yoffset[workerId])) / canvasHeight,
                        -1, 1 - (2.f * (x + xoffset[workerId])) / canvasWidth);
                d.x /= aspect;
                d.z *= aspect;
                d.normalize();
                bufferedTransform.transform(d);
                
                Vector3f oo = vectorPool[workerId].get(o);
                if (bufferedRenderSlow)
                	octree.trace2(d, oo,
                	        light, 0.8f, 0.2f, buf0, buf1, buf2,
                	        rayPool[workerId], vectorPool[workerId]);
                else
                	octree.quickTrace(d, oo,
                	        buf0, rayPool[workerId]);
                vectorPool[workerId].dispose(oo);
                vectorPool[workerId].dispose(d);
                	
                buf0[0] = Math.min(1.f, buf0[0]);
                buf0[1] = Math.min(1.f, buf0[1]);
                buf0[2] = Math.min(1.f, buf0[2]);
                buffer.setRGB(x, y, Color.getRGB(buf0[0], buf0[1], buf0[2]));
            }
        }
        vectorPool[workerId].dispose(o);
        vectorPool[workerId].dispose(light);
    }

	static Matrix3f tmpTransform = new Matrix3f();
	
	/**
	 * Update the camera and world transformation matrices.
	 */
	private synchronized void updateTransforms() {
	    tmpTransform.rotZ(pitch);
        cameraTransform.rotY(yaw);
	    cameraTransform.mul(tmpTransform);
        
        refresh = true;
        notifyAll();
	}

	public synchronized void renderSlow() {
		if (renderSlow)
            return;
		renderSlow = true;
        refresh = true;
        notifyAll();
	}
	
	public synchronized void moveForward(float v) {
	    Vector3f d = vectorPool[NUM_WORKERS].get(0, -1, 0);
	    cameraTransform.transform(d);
	    cameraPos.scaleAdd(v, d, cameraPos);
	    refresh = true;
	    notifyAll();
	    vectorPool[NUM_WORKERS].dispose(d);
	}
	
	public synchronized void moveBackward(float v) {
	    Vector3f d = vectorPool[NUM_WORKERS].get(0, -1, 0);
        cameraTransform.transform(d);
	    cameraPos.scaleAdd(-v, d, cameraPos);
        refresh = true;
        notifyAll();
        vectorPool[NUM_WORKERS].dispose(d);
    }
	
	public synchronized void strafeLeft(float v) {
	    Vector3f d = vectorPool[NUM_WORKERS].get(0, -1, 0);
        cameraTransform.transform(d);
	    Vector3f right = vectorPool[NUM_WORKERS].get();
	    right.cross(d, up);
	    cameraPos.scaleAdd(-v, right, cameraPos);
        refresh = true;
        notifyAll();
        vectorPool[NUM_WORKERS].dispose(right);
        vectorPool[NUM_WORKERS].dispose(d);
    }
	
	public synchronized void strafeRight(float v) {
	    Vector3f d = vectorPool[NUM_WORKERS].get(0, -1, 0);
        cameraTransform.transform(d);
	    Vector3f right = vectorPool[NUM_WORKERS].get();
        right.cross(d, up);
        cameraPos.scaleAdd(v, right, cameraPos);
        refresh = true;
        notifyAll();
        vectorPool[NUM_WORKERS].dispose(right);
        vectorPool[NUM_WORKERS].dispose(d);
    }
	
	public synchronized void rotateView(float yaw, float pitch) {
	    final float twopi = (float) (Math.PI * 2);
	    this.yaw += yaw;
	    this.pitch += pitch;
	    
	    this.pitch = Math.min(0, this.pitch);
	    this.pitch = Math.max(- (float) Math.PI, this.pitch);
	    
	    if (this.yaw > twopi)
	        this.yaw -= twopi;
	    else if (this.yaw < -twopi)
	        this.yaw += twopi;
	    
	    updateTransforms();
	}

}
