/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.octree;

import java.util.Stack;

import javax.vecmath.Vector3f;

import se.llbit.chunky.meta.Color;
import se.llbit.chunky.meta.Ray;
import se.llbit.chunky.meta.Ray.RayPool;
import se.llbit.chunky.meta.VectorPool;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.BlockType;

/**
 * Octree. Only thread safe if you finish twiddling with it before
 * you start tracing rays in it!
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class Octree {
	
	private class OctNode {
		public int type;
		public OctNode[] children;
		
		public OctNode(int type) {
			this.type = type;
		}

		public void subdivide() {
			children = new OctNode[8];
			children[0] = new OctNode(type);
			children[1] = new OctNode(type);
			children[2] = new OctNode(type);
			children[3] = new OctNode(type);
			children[4] = new OctNode(type);
			children[5] = new OctNode(type);
			children[6] = new OctNode(type);
			children[7] = new OctNode(type);
			type = -1;
		}
		
		public void merge(int newType) {
			type = newType;
			children = null;
			
		}
	}
	
	private int levels;
	private OctNode root;
	private OctNode cache;
	private int cx = 0;
	private int cy = 0;
	private int cz = 0;
	private int cacheLevel;
	
	/**
	 * Create a new Octree. The maximum voxel dimensions of the Octree are 2^levels.
	 * @param levels The number of levels in the Octree.
	 */
	public Octree(int levels) {
		this.levels = levels;
		root = new OctNode(0);
		flushCache();
	}
	
	private Stack<OctNode> parents = new Stack<OctNode>();
	
	/**
	 * Set the voxel type at the given coordinates.
	 * 
	 * @param type The new voxel type to be set
	 * @param x
	 * @param y
	 * @param z
	 */
	public synchronized void set(int type, int x, int y, int z) {
	    parents.clear();
		OctNode node = root;
		for (int i = levels-1; i >= 0; --i) {
			parents.add(node);

			if (node.type == type) {
				return;
			} else if (node.type != -1) {
				node.subdivide();
				parents.clear();
			}
			
			int xbit = 0x01 & (x >> i);
			int ybit = 0x01 & (y >> i);
			int zbit = 0x01 & (z >> i);
			node = node.children[(xbit<<2) | (ybit<<1) | zbit];
			
		}
		node.type = type;
		
		// apply changes
		int parentLevel = 0;
		for (int i = parents.size()-1; i >= 0; --i, ++parentLevel) {
			OctNode parent = parents.pop();
			
			boolean allSame = true;
			for (OctNode child: parent.children) {
				if (child.type != node.type) {
					allSame = false;
					break;
				}
			}

			if (allSame) {
				parent.merge(node.type);
				if (parentLevel > cacheLevel)
					flushCache();
			} else {
				break;
			}
		}
		
	}
	
	/**
	 * @param x
	 * @param y
	 * @param z
	 * @return The voxel type at the given coordinates
	 */
	public synchronized int get(int x, int y, int z) {
		if (cache != root && ((x >>> cacheLevel) != cx ||
				(y >>> cacheLevel) != cy ||
				(z >>> cacheLevel) != cz))
			flushCache();
		
		int level = cacheLevel;
		for (; level >= 0; --level) {
			if (cache.type != -1)
				break;
			
			int xbit = 0x01 & (x >> level);
			int ybit = 0x01 & (y >> level);
			int zbit = 0x01 & (z >> level);
			cache = cache.children[(xbit<<2) | (ybit<<1) | zbit];
			cacheLevel = level;
		}
		cx = x >>> cacheLevel;
		cy = y >>> cacheLevel;
		cz = z >>> cacheLevel;
		return cache.type;
	}
	
	private synchronized void flushCache() {
		cache = root;
		cacheLevel = levels-1;
		cx = 0;
		cy = 0;
		cz = 0;
	}

	private static final float EPSILON = 0.000005f;
	private static final float RAY_OFFSET = 0.0001f;
	
    /**
     * Test whether the ray intersects any voxel before exiting the Octree.
     * @param ray the ray
     * @return <code>true</code> if the ray intersects a voxel
     */
    public boolean intersect(Ray ray) {

        while (true) {
        	Vector3f o = ray.o;
        	Vector3f d = ray.d;
        	
            int x = (int) Math.floor(o.x);
            int y = (int) Math.floor(o.y);
            int z = (int) Math.floor(o.z);
            
            if (Math.abs(x) >= (1<<levels-1)
                    || Math.abs(y) >= (1<<levels-1)
                    || Math.abs(z) >= (1<<levels-1) ) {
                ray.currentMaterial = Block.AIR.id;
                return false;// outside of octree!
            }
            
            OctNode node = root;
            int level = levels-1;
            for (int i = level; node.type == -1 && i >= 0; --i) {
                level = i;
                int xbit = 0x01 & (x >> level);
                int ybit = 0x01 & (y >> level);
                int zbit = 0x01 & (z >> level);
                node = node.children[(xbit<<2) | (ybit<<1) | zbit];
            }
            
            if (ray.currentMaterial == -1) {
            	ray.currentMaterial = node.type;
            	
            } else if (node.type != ray.currentMaterial) {
            	ray.currentMaterial = node.type;
                return true;
            }
            
            int nx = x >>> level;
            int ny = y >>> level;
            int nz = z >>> level;
            
            float tNear = Float.POSITIVE_INFINITY;
            float t1, t2;
            if (d.x != 0) {
                t1 = ((nx<<level) - o.x) / d.x;
                t2 = (((nx+1)<<level) - o.x) / d.x;
                if (t1 > EPSILON && t1 < tNear) {
                    tNear = t1;
                    ray.intersectionNormal.x = -1;
                    ray.intersectionNormal.y = 0;
                    ray.intersectionNormal.z = 0;
                }
                if (t2 > EPSILON && t2 < tNear) {
                    tNear = t2;
    	            ray.intersectionNormal.x = 1;
    	            ray.intersectionNormal.y = 0;
    	            ray.intersectionNormal.z = 0;
    	        }
            }
            if (d.y != 0) {
                t1 = ((ny<<level) - o.y) / d.y;
                t2 = (((ny+1)<<level) - o.y) / d.y;
                if (t1 > EPSILON && t1 < tNear) {
                    tNear = t1;
    	            ray.intersectionNormal.x = 0;
    	            ray.intersectionNormal.y = -1;
    	            ray.intersectionNormal.z = 0;
    	        }
                if (t2 > EPSILON && t2 < tNear) {
                    tNear = t2;
    	            ray.intersectionNormal.x = 0;
    	            ray.intersectionNormal.y = 1;
    	            ray.intersectionNormal.z = 0;
    	        }
            }
            if (d.z != 0) {
                t1 = ((nz<<level) - o.z) / d.z;
                t2 = (((nz+1)<<level) - o.z) / d.z;
                if (t1 > EPSILON && t1 < tNear) {
                    tNear = t1;
    	            ray.intersectionNormal.x = 0;
    	            ray.intersectionNormal.y = 0;
    	            ray.intersectionNormal.z = -1;
    	        }
                if (t2 > EPSILON && t2 < tNear) {
                    tNear = t2;
    	            ray.intersectionNormal.x = 0;
    	            ray.intersectionNormal.y = 0;
    	            ray.intersectionNormal.z = 1;
    	        }
            }
            
            ray.intersection.x = o.x + d.x * tNear;
            ray.intersection.y = o.y + d.y * tNear;
            ray.intersection.z = o.z + d.z * tNear;
            
            // add small offset past the intersection to avoid
            // recursion to the same octree node!
            tNear += RAY_OFFSET;
            o.x = o.x + d.x * tNear;
            o.y = o.y + d.y * tNear;
            o.z = o.z + d.z * tNear;
        }
    }
    
    /**
     * @param d
     * @param o
     * @return color
     */
    public void quickTrace(Vector3f d, Vector3f o, float[] color, RayPool queue) {
    	
        Ray ray = queue.get();
    	ray.d = d;
    	ray.o = o;
    	
    	intersect(ray);
    	
    	if (ray.currentMaterial != 0) {
    	    //return getBorderColor(ray);
    	    Block block = Block.values[ray.currentMaterial];
    	    Color.getRGBAComponents(block.getAvgRGB(), color);
    	} else {
    	
    	    color[0] = color[1] = color[2] = .5f;
    	}
    	
    	queue.dispose(ray);
    }
    
    /**
     * Whitted ray tracing.
     * @param d ray direction
     * @param o origin (previous hit location)
     * @param light light position
     * @param lightIntensity 
     * @param ambientIntensity 
     */
    public void trace2(Vector3f d, Vector3f o,
            Vector3f light, float lightIntensity, float ambientIntensity,
            float[] color, float[] buf1, float[] buf2, RayPool pool,
            VectorPool vectorPool) {
    	
    	Ray ray = pool.get();
    	ray.d = d;
    	ray.o = o;
    	
    	rayTrace(ray, light, lightIntensity, ambientIntensity, buf1, buf2, pool,
    	        vectorPool);
    	
    	color[0] = ray.color[0];
        color[1] = ray.color[1];
        color[2] = ray.color[2];
        pool.dispose(ray);
    }
    
    /**
     * @param ray
     * @param light
     * @param lightIntensity
     * @param ambientIntensity
     */
    public void rayTrace(Ray ray, Vector3f light,
            float lightIntensity, float ambientIntensity,
            float[] color, float[] shadowColor, RayPool pool,
            VectorPool vectorPool) {
    	int m1 = Block.AIR.id;
    	
    	float tNear;
        float t1, t2;
    	Ray transmittedRay;
    	
    	while (intersect(ray)) {
    		
    	    switch (Block.values[ray.currentMaterial]) {
    	    case AIR:
    	        break;
    	    case LILYPAD:
    	        
                int x = (int) Math.floor(ray.o.x);
                int y = (int) Math.floor(ray.o.y);
                int z = (int) Math.floor(ray.o.z);
                
                tNear = Float.POSITIVE_INFINITY;
                Vector3f d = ray.d;
                if (d.x != 0) {
                    t1 = (x - ray.o.x) / d.x;
                    t2 = ((x+1) - ray.o.x) / d.x;
                    if (t1 > EPSILON && t1 < tNear) {
                        tNear = t1;
                        ray.intersectionNormal.x = -1;
                        ray.intersectionNormal.y = 0;
                        ray.intersectionNormal.z = 0;
                    }
                    if (t2 > EPSILON && t2 < tNear) {
                        tNear = t2;
                        ray.intersectionNormal.x = 1;
                        ray.intersectionNormal.y = 0;
                        ray.intersectionNormal.z = 0;
                    }
                }
                if (d.z != 0) {
                    t1 = (z - ray.o.z) / d.z;
                    t2 = ((z+1) - ray.o.z) / d.z;
                    if (t1 > EPSILON && t1 < tNear) {
                        tNear = t1;
                        ray.intersectionNormal.x = 0;
                        ray.intersectionNormal.y = 0;
                        ray.intersectionNormal.z = -1;
                    }
                    if (t2 > EPSILON && t2 < tNear) {
                        tNear = t2;
                        ray.intersectionNormal.x = 0;
                        ray.intersectionNormal.y = 0;
                        ray.intersectionNormal.z = 1;
                    }
                }
                boolean hit = false;
                if (d.y != 0) {
                    t1 = (y - ray.o.y) / d.y;
                    if (t1 > EPSILON && t1 < tNear) {
                        hit = true;
                        tNear = t1;
                        ray.intersectionNormal.x = 0;
                        ray.intersectionNormal.y = -1;
                        ray.intersectionNormal.z = 0;
                    }
                }
                
                transmittedRay = pool.get(ray);
                transmittedRay.o.set(ray.o);
                transmittedRay.d.set(ray.d);
                
                ray.intersection.x = ray.o.x + d.x * tNear;
                ray.intersection.y = ray.o.y + d.y * tNear;
                ray.intersection.z = ray.o.z + d.z * tNear;
                
                // add small offset past the intersection to avoid
                // recursion to the same octree node!
                tNear += RAY_OFFSET;
                ray.o.x = ray.o.x + d.x * tNear;
                ray.o.y = ray.o.y + d.y * tNear;
                ray.o.z = ray.o.z + d.z * tNear;
                    
                if (hit) {
                    getIntersectionColor(ray, color);
                    
                    transmittedRay.attenuation = ray.attenuation * (1 - color[3]);
                    ray.attenuation *= color[3];
                    
                    if (ray.attenuation > EPSILON) {
                        Vector3f lightDir = vectorPool.get(light);
                        lightDir.sub(ray.intersection);
                        lightDir.normalize();
                        
                        Ray shadowRay = pool.get(ray);
                        shadowRay.o.set(lightDir);
                        shadowRay.o.scaleAdd(0.01f, ray.intersection);
                        shadowRay.d.set(lightDir);
                        shadowRay.currentMaterial = m1;
                        
                        float intensity = 0.f;
                        while (intersect(shadowRay)) {
                            getIntersectionColor(shadowRay, shadowColor);
                            shadowRay.attenuation *= (1 - shadowColor[3]);
                        }
                        if (shadowRay.attenuation > EPSILON) {
                            lightDir.scale(-1);
                            intensity = shadowRay.attenuation * lightIntensity * ray.intersectionNormal.dot(lightDir);
                            intensity = Math.max(0.f, intensity);
                        }
                        vectorPool.dispose(lightDir);
                        pool.dispose(shadowRay);
                        intensity += ambientIntensity;
                        ray.attenuation *= intensity;
                        
                        ray.color[0] = color[0] * ray.attenuation;
                        ray.color[1] = color[1] * ray.attenuation;
                        ray.color[2] = color[2] * ray.attenuation;
                    }
                    
                    if (transmittedRay.attenuation > EPSILON && transmittedRay.rayDepth > 0) {
                        
                        // water refraction
                        if (BlockType.water.contains(m1) || BlockType.water.contains(ray.currentMaterial)) {
                            float n1 = Block.values[m1].refractionIndex;
                            float n2 = Block.values[ray.currentMaterial].refractionIndex;
                            float n1n2 = n1 / n2;
                            Vector3f minusd = vectorPool.get(-ray.d.x, -ray.d.y, -ray.d.z);
                            t1 = ray.intersectionNormal.dot(minusd);
                            vectorPool.dispose(minusd);
                            t2 = (float) Math.sqrt(1 - n1n2*n1n2 * (1 - t1*t1));
                            if (t1 > 0) {
                                transmittedRay.d.x = n1n2*ray.d.x + (n1n2*t1 - t2)*ray.intersectionNormal.x;
                                transmittedRay.d.y = n1n2*ray.d.y + (n1n2*t1 - t2)*ray.intersectionNormal.y;
                                transmittedRay.d.z = n1n2*ray.d.z + (n1n2*t1 - t2)*ray.intersectionNormal.z;
                            } else {
                                transmittedRay.d.x = n1n2*ray.d.x - (n1n2*t1 - t2)*ray.intersectionNormal.x;
                                transmittedRay.d.y = n1n2*ray.d.y - (n1n2*t1 - t2)*ray.intersectionNormal.y;
                                transmittedRay.d.z = n1n2*ray.d.z - (n1n2*t1 - t2)*ray.intersectionNormal.z;
                            }
                            transmittedRay.o.x = ray.intersection.x + transmittedRay.d.x * RAY_OFFSET;
                            transmittedRay.o.y = ray.intersection.y + transmittedRay.d.y * RAY_OFFSET;
                            transmittedRay.o.z = ray.intersection.z + transmittedRay.d.z * RAY_OFFSET;
                        }
                        
                        rayTrace(transmittedRay, light, lightIntensity, ambientIntensity, color, shadowColor, pool, vectorPool);
                        if (ray.currentMaterial != 0) {
                            ray.color[0] += transmittedRay.color[0];
                            ray.color[1] += transmittedRay.color[1];
                            ray.color[2] += transmittedRay.color[2];
                        }
                    }
                    pool.dispose(transmittedRay);
                    return;
                }
                
    	        break;
    	        
    	    default:
    			getIntersectionColor(ray, color);

    			transmittedRay = pool.get(ray);
    			transmittedRay.o.set(ray.o);
    			transmittedRay.d.set(ray.d);
    			
    			Ray reflectedRay = null;
    			
    			if (ray.currentMaterial == Block.WATER.id) {
    				final float reflectance = 0.4f;
	    			reflectedRay = pool.get(ray);
	    			reflectedRay.o.set(ray.o);
	    			reflectedRay.d.scaleAdd(
	    			        - 2 * ray.d.dot(ray.intersectionNormal),
	    			        ray.intersectionNormal, ray.d);
	    			reflectedRay.attenuation = ray.attenuation * reflectance;
	    			ray.attenuation *= (1 - reflectance);
    			}
    			
    			transmittedRay.attenuation = ray.attenuation * (1 - color[3]);
    			ray.attenuation *= color[3];
    			
    			if (ray.attenuation > EPSILON) {
    			    Ray shadowRay = pool.get(ray);
    			    shadowRay.d.set(light);
    			    shadowRay.d.sub(ray.intersection);
    			    shadowRay.d.normalize();
    				shadowRay.o.set(shadowRay.d);
    				shadowRay.o.scaleAdd(0.01f, ray.intersection);
    				
    				shadowRay.currentMaterial = m1;
    				
    				float intensity = 0.f;
    				while (intersect(shadowRay)) {
    					getIntersectionColor(shadowRay, shadowColor);
    					shadowRay.attenuation *= (1 - shadowColor[3]);
    				}
    				if (shadowRay.attenuation > EPSILON) {
    					intensity = shadowRay.attenuation * lightIntensity *
    					        (- ray.intersectionNormal.dot(shadowRay.d));
    					intensity = Math.max(0.f, intensity);
    				}
    				pool.dispose(shadowRay);
    				intensity += ambientIntensity;
    				ray.attenuation *= intensity;
    				
    				ray.color[0] = color[0] * ray.attenuation;
                    ray.color[1] = color[1] * ray.attenuation;
                    ray.color[2] = color[2] * ray.attenuation;
    				
    				if (ray.currentMaterial == Block.WATER.id && reflectedRay.rayDepth > 0) {
    					rayTrace(reflectedRay, light, lightIntensity, ambientIntensity, color, shadowColor, pool, vectorPool);
    					ray.color[0] += reflectedRay.color[0];
                        ray.color[1] += reflectedRay.color[1];
                        ray.color[2] += reflectedRay.color[2];
                        pool.dispose(reflectedRay);
    				}
    			}
    			
    			if (transmittedRay.attenuation > EPSILON && transmittedRay.rayDepth > 0) {
    				
    				// water refraction
    	            if (BlockType.water.contains(m1) || BlockType.water.contains(ray.currentMaterial)) {
    	                float n1 = Block.values[m1].refractionIndex;
    	                float n2 = Block.values[ray.currentMaterial].refractionIndex;
    	                float n1n2 = n1 / n2;
    	                Vector3f minusd = vectorPool.get(-ray.d.x, -ray.d.y, -ray.d.z);
    	                t1 = ray.intersectionNormal.dot(minusd);
    	                vectorPool.dispose(minusd);
    	                t2 = (float) Math.sqrt(1 - n1n2*n1n2 * (1 - t1*t1));
    	                if (t1 > 0) {
	    	                transmittedRay.d.x = n1n2*ray.d.x + (n1n2*t1 - t2)*ray.intersectionNormal.x;
	    	                transmittedRay.d.y = n1n2*ray.d.y + (n1n2*t1 - t2)*ray.intersectionNormal.y;
	    	                transmittedRay.d.z = n1n2*ray.d.z + (n1n2*t1 - t2)*ray.intersectionNormal.z;
    	                } else {
    	                	transmittedRay.d.x = n1n2*ray.d.x - (n1n2*t1 - t2)*ray.intersectionNormal.x;
	    	                transmittedRay.d.y = n1n2*ray.d.y - (n1n2*t1 - t2)*ray.intersectionNormal.y;
	    	                transmittedRay.d.z = n1n2*ray.d.z - (n1n2*t1 - t2)*ray.intersectionNormal.z;
    	                }
    	                transmittedRay.o.x = ray.intersection.x + transmittedRay.d.x * RAY_OFFSET;
    	                transmittedRay.o.y = ray.intersection.y + transmittedRay.d.y * RAY_OFFSET;
    	                transmittedRay.o.z = ray.intersection.z + transmittedRay.d.z * RAY_OFFSET;
    	            }
    				
    	    		rayTrace(transmittedRay, light, lightIntensity, ambientIntensity, color, shadowColor, pool, vectorPool);
    	    		ray.color[0] += transmittedRay.color[0];
                    ray.color[1] += transmittedRay.color[1];
                    ray.color[2] += transmittedRay.color[2];
    	    	}
    			pool.dispose(transmittedRay);
    			return;
        	}
    		
    		m1 = ray.currentMaterial;
    		
    	}

    	// sky color
    	ray.color[0] = ray.color[1] = ray.color[2] = ray.attenuation * .5f;
    }
    
    private void getBorderColor(Ray ray, float[] color) {
        int x = (int) Math.floor(ray.intersection.x);
        int y = (int) Math.floor(ray.intersection.y);
        int z = (int) Math.floor(ray.intersection.z);
        
        float u, v;
        if (ray.intersectionNormal.y != 0) {
            u = ray.intersection.x - x;
            v = ray.intersection.z - z;
        } else if (ray.intersectionNormal.x != 0) {
            v = 1 - (ray.intersection.y - y);
            u = (ray.intersection.z - z);
        } else {
            u = ray.intersection.x - x;
            v = 1 - (ray.intersection.y - y);
        }
        float m = Math.max(Math.abs(.5f - u), Math.abs(.5f - v));
        color[0] = color[1] = color[2] = m;
    }

	private float[] getIntersectionColor(Ray ray, float[] color) {
		int x = (int) Math.floor(ray.intersection.x);
        int y = (int) Math.floor(ray.intersection.y);
        int z = (int) Math.floor(ray.intersection.z);
        
		float u, v;
		if (ray.intersectionNormal.y != 0) {
			u = ray.intersection.x - x;
			v = ray.intersection.z - z;
		} else if (ray.intersectionNormal.x != 0) {
			v = 1 - (ray.intersection.y - y);
			u = (ray.intersection.z - z);
		} else {
			u = ray.intersection.x - x;
			v = 1 - (ray.intersection.y - y);
		}
		Block block = Block.values[ray.currentMaterial];
		int rgb = block.getColor(u, v);
		color[0] = (0xFF & (rgb >> 16)) / 255.f;
		color[1] = (0xFF & (rgb >> 8)) / 255.f;
		color[2] = (0xFF & rgb) / 255.f;
		color[3] = (0xFF & (rgb >> 24)) / 255.f;
		return color;
	}

}
