/* Copyright (c) 2010 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.quadtree;

import java.util.ArrayList;
import java.util.List;

public class Quadtree {

	private class QuadNode {
		public int value;
		public QuadNode[] children;
		
		public QuadNode(int type) {
			this.value = type;
		}

		public void subdivide() {
			children = new QuadNode[4];
			children[0] = new QuadNode(value);
			children[1] = new QuadNode(value);
			children[2] = new QuadNode(value);
			children[3] = new QuadNode(value);
			value = -1;
		}
		
		public void merge(int newType) {
			value = newType;
			children = null;
			
		}
	}
	
	public int levels;
	private QuadNode root;
	private QuadNode cache;
	private int cx = 0;
	private int cy = 0;
	private int cacheLevel;
	
	public Quadtree(int nlevels) {
		levels = nlevels;
		root = new QuadNode(0);
		flushCache();
	}
	
	public Quadtree(int nlevels, int initial) {
		this(nlevels);
		root.value = initial;
	}

	public synchronized void set(int value, int x, int y) {
		List<QuadNode> parents = new ArrayList<QuadNode>();
		QuadNode node = root;
		for (int i = levels-1; i >= 0; --i) {
			parents.add(node);

			if (node.value == value) {
				return;
			} else if (node.value != -1) {
				node.subdivide();
				parents.clear();
			}
			
			int xbit = 0x01 & (x >> i);
			int ybit = 0x01 & (y >> i);
			node = node.children[(xbit<<1) | ybit];
			
		}
		node.value = value;
		
		// apply changes
		int parentLevel = 0;
		for (int i = parents.size()-1; i >= 0; --i, ++parentLevel) {
			QuadNode parent = parents.get(i);
			
			boolean allSame = true;
			for (QuadNode child: parent.children) {
				if (child.value != node.value) {
					allSame = false;
					break;
				}
			}

			if (allSame) {
				parent.merge(node.value);
				if (parentLevel > cacheLevel)
					flushCache();
			} else {
				break;
			}
		}
		
	}
	
	public synchronized int get(int x, int y) {
		if (cache != root && ((x >>> cacheLevel) != cx ||
				(y >>> cacheLevel) != cy))
			flushCache();
		
		int level = cacheLevel;
		for (; level >= 0; --level) {
			if (cache.value != -1)
				break;
			
			int xbit = 0x01 & (x >> level);
			int ybit = 0x01 & (y >> level);
			cache = cache.children[(xbit<<1) | ybit];
			cacheLevel = level;
		}
		cx = x >>> cacheLevel;
		cy = y >>> cacheLevel;
		return cache.value;
	}
	
	private synchronized void flushCache() {
		cache = root;
		cacheLevel = levels-1;
		cx = 0;
		cy = 0;
	}
}
